module myglobals
contains
  !
  ! Broadcasts 'count' elements of 'data' from rank 'root' to all
  ! other processes in the communicator 'comm'.
  !
  subroutine bcast_Int(data, count, root, comm)
    implicit none
    integer, dimension(*), intent(inout) :: data
    integer, intent(in) :: count, root, comm
    ! TODO: Implement using only send & receive calls
  end subroutine bcast_Int

  !
  ! Scatters the content of 'sendbuf' in process with rank 'root' to
  ! all processes in the communicator 'comm'. Each piece of data is
  ! 'sendcnt' elements long. Received data goes into 'recvbuf'.
  !
  subroutine scatter_Int(sendbuf, sendcnt, recvbuf, recvcnt, root, comm)
    implicit none
    integer, dimension(*), intent(in) :: sendbuf
    integer, intent(in) :: sendcnt
    integer, dimension(*), intent(out) :: recvbuf
    integer, intent(in) :: recvcnt, root, comm
    ! TODO: Implement using only send & receive calls
  end subroutine scatter_Int

  !
  ! Gathers the content of each process' 'sendbuf' into the 'recvbuf'
  ! of process with rank 'root'. Each process sends 'sendcount' pieces
  ! of data and the root receives 'recvcnt' from each process.
  !
  subroutine gather_Int(sendbuf, sendcnt, recvbuf, recvcnt, root, comm)
    implicit none
    integer, dimension(*), intent(in) :: sendbuf
    integer, intent(in) :: sendcnt
    integer, dimension(*), intent(out) :: recvbuf
    integer, intent(in) :: recvcnt, root, comm
    ! TODO: Implement using only send & receive calls
  end subroutine gather_Int

  !
  ! Scatters the content of each process' 'sendbuf' into the 'recvbuf'
  ! of each other process. Received pieces must be ordered by their
  ! sender's rank.
  !
  subroutine alltoall_Int(sendbuf, sendcnt, recvbuf, recvcnt, comm)
    implicit none
    integer, dimension(*), intent(in) :: sendbuf
    integer, intent(in) :: sendcnt
    integer, dimension(*), intent(out) :: recvbuf
    integer, intent(in) :: recvcnt, comm
    ! TODO: Implement using only send & receive calls
  end subroutine alltoall_Int

  !
  ! Performs a global summation reduction over the single integer value
  ! in all processes' 'sendbuf'-s. The result should be available in
  ! 'recvbuf' only at process with rank 'root'.
  !
  subroutine sum_Int(sendbuf, recvbuf, root, comm)
    implicit none
    integer, intent(in) :: sendbuf
    integer, intent(out) :: recvbuf
    integer, intent(in) :: root, comm
    ! TODO: Implement using only send & receive calls
  end subroutine sum_Int

end module myglobals

program global
  use ifport
  use mpi
  use myglobals
  implicit none
  integer :: myRank, numProcs, ierr
  integer :: i, myRandomNumber
  integer :: inspectRank = 0
  integer, dimension(:), allocatable :: mySendBuffer, myRecvBuffer
  integer, dimension(1) :: myData
  character(len=128) :: cmdArg

  call MPI_INIT(ierr)
  call MPI_COMM_SIZE(MPI_COMM_WORLD, numProcs, ierr)
  call MPI_COMM_RANK(MPI_COMM_WORLD, myRank, ierr)

  ! Allocate data buffers
  allocate(mySendBuffer(numProcs))
  allocate(myRecvBuffer(numProcs))

  ! Seed the random number generator with a rank-specific value
  call srand(int(MPI_Wtime()+123*myRank))
  myRandomNumber = mod(irand(), 100)

  if (iargc() > 0) then
     call getarg(1, cmdArg)
     read(cmdArg,*) inspectRank
     if (inspectRank < 0 .or. inspectRank >= numProcs) &
          call MPI_ABORT(MPI_COMM_WORLD, 1, ierr)
  end if

  print '("Rank ",I2.1," has the random number ",I0)', myRank, myRandomNumber

  myData = -1

  ! Use MPI barrier to synchronize the output
  call MPI_BARRIER(MPI_COMM_WORLD, ierr)

  ! ---- Broadcast operation ----

  if (myRank == 0) print '(A,/,A,/,A)', &
       "-------------------------------------", &
       "broadcast", &
       "-------------------------------------"

  ! Prepare the local data
  myData = myRandomNumber

  if (myRank == inspectRank) print '(A,I2.1,A,I0)', &
       "Rank ", inspectRank, ": myData before the broadcast is ", myData

  ! Call the user's implementation of broadcast
  call bcast_Int(myData, 1, 0, MPI_COMM_WORLD)
  if (myRank == inspectRank) print '(A,I2.1,A,I0)', &
       "Rank ", inspectRank, ": myData after the broadcast is ", myData

  ! TODO: Insert a MPI broadcast call here
  if (myRank == inspectRank) print '(A,I2.1,A,I0)', &
       "Rank ", inspectRank, ": myData after the MPI broadcast is ", myData

  call MPI_BARRIER(MPI_COMM_WORLD, ierr)

  ! ---- Scatter operation ----

  if (myRank == 0) print '(A,/,A,/,A)', &
       "-------------------------------------", &
       "scatter", &
       "-------------------------------------"

  ! Prepare the local data. Mark each random number with a prefix
  ! accord to its position in the send buffer
  mySendBuffer = (/(myRandomNumber + i*100, i = 0, numProcs-1)/)
  myData = -1

  if (myRank == inspectRank) print '(A,I2.1,A,I0)', &
       "Rank ", inspectRank, ": myData before the scatter is ", myData

  ! Call the user's implementation of scatter
  call scatter_Int(mySendBuffer, 1, myData, 1, 0, MPI_COMM_WORLD)
  if (myRank == inspectRank) print '(A,I2.1,A,I0)', &
       "Rank ", inspectRank, ": myData after the scatter is ", myData

  ! TODO: Insert a MPI scatter call here
  if (myRank == inspectRank) print '(A,I2.1,A,I0)', &
       "Rank ", inspectRank, ": myData after the MPI scatter is ", myData

  call MPI_BARRIER(MPI_COMM_WORLD, ierr)

  ! ---- Gather operation ----

  if (myRank == 0) print '(A,/,A,/,A)', &
       "-------------------------------------", &
       "gather", &
       "-------------------------------------"

  ! Prepare the local data
  myRecvBuffer = -1
  myData = myRandomNumber

  if (myRank == inspectRank) print '(A,I2.1,A,<numProcs>(" ",I0))', &
       "Rank ", inspectRank, ": myRecvBuffer before the gather is", &
       (myRecvBuffer(i), i=1,numProcs)

  ! Call the user's implementation of gather
  call gather_Int(myData, 1, myRecvBuffer, 1, 0, MPI_COMM_WORLD)
  if (myRank == inspectRank) print '(A,I2.1,A,<numProcs>(" ",I0))', &
       "Rank ", inspectRank, ": myRecvBuffer after the gather is", &
       (myRecvBuffer(i), i=1,numProcs)

  ! TODO: Insert a MPI gather call here
  if (myRank == inspectRank) print '(A,I2.1,A,<numProcs>(" ",I0))', &
       "Rank ", inspectRank, ": myRecvBuffer after the MPI gather is", &
       (myRecvBuffer(i), i=1,numProcs)

  call MPI_BARRIER(MPI_COMM_WORLD, ierr)

  ! ---- All-to-all operation ----

  if (myRank == 0) print '(A,/,A,/,A)', &
       "-------------------------------------", &
       "all-to-all", &
       "-------------------------------------"

  ! Prepare the local data
  mySendBuffer = (/(myRandomNumber + i*100, i = 0, numProcs-1)/)
  myRecvBuffer = -1

  if (myRank == inspectRank) print '(A,I2.1,A,<numProcs>(" ",I0))', &
       "Rank ", inspectRank, ": myRecvBuffer before the all-to-all  is", &
       (myRecvBuffer(i), i=1,numProcs)

  ! Call the user's implementation of all-to-all
  call alltoall_Int(mySendBuffer, 1, myRecvBuffer, 1, MPI_COMM_WORLD)
  if (myRank == inspectRank) print '(A,I2.1,A,<numProcs>(" ",I0))', &
       "Rank ", inspectRank, ": myRecvBuffer after the all-to-all is", &
       (myRecvBuffer(i), i=1,numProcs)

  ! TODO: Insert a MPI all-to-all call here
  if (myRank == inspectRank) print '(A,I2.1,A,<numProcs>(" ",I0))', &
       "Rank ", inspectRank, ": myRecvBuffer after the MPI all-to-all is", &
       (myRecvBuffer(i), i=1,numProcs)

  call MPI_BARRIER(MPI_COMM_WORLD, ierr)

  ! ---- Gloal reduction operation ----

  if (myRank == 0) print '(A,/,A,/,A)', &
       "-------------------------------------", &
       "reduce", &
       "-------------------------------------"

  ! Prepare the local data
  myData = -1

  if (myRank == inspectRank) print '(A,I2.1,A,I0)', &
       "Rank ", inspectRank, ": myData before the reduction is ", myData

  ! Call the user's implementation of global reduction to sum all ranks
  call sum_Int(myRank, myData(1), 0, MPI_COMM_WORLD)
  if (myRank == inspectRank) print '(A,I2.1,A,I0)', &
       "Rank ", inspectRank, ": myData after the reduction is ", myData

  ! TODO: Insert a MPI reduction call here
  if (myRank == inspectRank) print '(A,I2.1,A,I0)', &
       "Rank ", inspectRank, ": myData after the MPI reduction is ", myData

  call MPI_FINALIZE(ierr)

  print '("Former rank ",I2.1," about to shut down")', myRank
end program global
